//=============================================================================
//
// BSD 3-Clause License
//
// Copyright © 2013-2015, DAVID 3D Solutions GbR
// All rights reserved.
//
// Redistribution and use in source and binary forms, 
// with or without modification, are permitted provided that 
// the following conditions are met:
//
// - Redistributions of source code must retain the above copyright notice, 
//   this list of conditions and the following disclaimer.
// - Redistributions in binary form must reproduce the above copyright notice, 
//   this list of conditions and the following disclaimer in the documentation 
//   and/or other materials provided with the distribution.
// - Neither the name of DAVID 3D Solutions GbR nor the names of its contributors 
//   may be used to endorse or promote products derived from this software 
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//=============================================================================

#include "davidSDK/TcpClient.h"
#include <iostream>
#include <string.h>

namespace david {

//=============================================================================
// TcpClient
//=============================================================================

TcpClient::TcpClient(const std::string& serverAddress, uint16_t serverPort)
	: m_serverAddress(serverAddress)
	, m_serverPort(serverPort)
	, m_socket(-1)
{
	// Init networking (Winsock DLL):
	WSADATA wsaData;
	WSAStartup(MAKEWORD(2, 2), &wsaData);
}


TcpClient::~TcpClient()
{
	if(m_socket != -1) this->Close();

	WSACleanup();
}


bool TcpClient::Connect()
{
    struct addrinfo *ptr = NULL;
    struct addrinfo hints;

	ZeroMemory(&hints, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	
	// get addressInfo about server:port:
    addrinfo *addressInfos = NULL;
	if (getaddrinfo(m_serverAddress.c_str(), std::to_string(m_serverPort).c_str(), &hints, &addressInfos) != 0)
	{
		std::cerr << "TcpClient::Connect(): getaddrinfo failed!\n";
	}
	
	// loop over all results in addressInfos:
	for(addrinfo* info=addressInfos; info!=NULL ;info=info->ai_next)
	{
		// try to create socket:
		SOCKET s = ::socket(info->ai_family, info->ai_socktype, info->ai_protocol);
		if (-1 == s) continue;

		// try to connect:
		if(-1 == ::connect(s, info->ai_addr, (int)info->ai_addrlen))
		{
			::closesocket(s);
			continue;
		}

		m_socket = s;
		break;
	}
	
	return (m_socket != -1);
}


void TcpClient::Close()
{
	if (m_socket != -1) ::closesocket(m_socket);
	m_socket = -1;
}


std::string TcpClient::GetServerAddress() const
{
	return m_serverAddress;
}


uint16_t TcpClient::GetServerPort() const
{
	return m_serverPort;
}


int64_t TcpClient::ReceiveString(std::string& data)
{
	char buffer[2000];
	int64_t bytesReceived = ::recv(m_socket, buffer, sizeof(buffer), 0);
	
	if (bytesReceived >= 0)  // no error
	{
		// copy received data:
		data = std::string(buffer, (size_t)bytesReceived);
	}
	else  // error
	{
		std::cerr << "TcpClient::ReceiveString(): Error receiving" << std::endl;
	}

	return bytesReceived;
}


bool TcpClient::ReceiveBinary(void* data, size_t numBytes)
{
	if(!data) return false;
		
	if(0 == numBytes) return true;
		
	size_t totalBytesReceived = 0;
	while(totalBytesReceived < numBytes)
	{
		int bytesLeft = (int)numBytes - (int)totalBytesReceived;
		int bytesReceived = ::recv(m_socket, ((char*)data)+totalBytesReceived, bytesLeft, 0);
		if(bytesReceived > 0)
		{
			totalBytesReceived += bytesReceived;
		}
		else if(bytesReceived < 0)  // error
		{
			std::cerr << "TcpClient::ReceiveBinary(): Error: not received enough bytes. Connection lost?" << std::endl;
			return false;
		}
		else  // bytesReceived == 0 --> connection is closed
		{
			break;
		}
	} 
		
	return (totalBytesReceived==numBytes);
}


int64_t TcpClient::SendString(const std::string& data)
{
	return ::send(m_socket, data.c_str(), (int)data.length(), 0);
}


bool TcpClient::SendBinary(const void* data, size_t numBytes)
{
	if(!data) 
	{
		std::cerr << "TcpClient::SendBinary(): Error: data pointer is NULL." << std::endl;
		return false;
	}
		
	// Send bytes:
	int bytesSent = ::send(m_socket, (const char*) data, (int)numBytes, MSG_DONTROUTE);
		
	// Check.
	if(bytesSent == (int)numBytes)
	{
		return true;
	}
	else
	{
		std::cerr << "TcpClient::SendBinary(): Error: send failed. Connection lost?" << std::endl;
		return false;
	}
}


} // namespace
